@@ -39,13 +39,13 @@ from tqdm import tqdm
 import matplotlib.pyplot as plt
 
 
-# %% Structure learning across all variables
+# %% Association learning across all variables
 class hnet():
     """HNET - Graphical Hypergeometric networks.
 
     Description
     -----------
-    This is the main function to detect significant edge probabilities between pairs of vertices (node-links) given the input data set **df**.
+    This is the main function to detect significant edge probabilities between pairs of vertices (node-links) given the input DataFrame.
 
     A multi-step process is performed which consisting 5 steps.
         1. Pre-processing: Typing and One-hot Enconding. Each feature is set as being categoric, numeric or is excluded. The typing can be user-defined or automatically determined on conditions. Encoding of features in a one-hot dense array is done for the categoric terms. The one-hot dense array is subsequently used to create combinatory features using k combinations over n features (without replacement).
@@ -59,59 +59,45 @@ class hnet():
 
     Parameters
     ----------
-    alpha : float [0..1], default : 0.05.
+    alpha : float [0..1], (default : 0.05)
         Significance to keep only edges with <=alhpa.
-        0.05 : (default)
         1 : (for all results)
-
-    y_min : int [1..n], where n is the number of samples. default : 10.
+    y_min : int [1..n], where n is the number of samples. (default : 10)
         Minimum number of samples in a group. Should be [samples>=y_min]. All groups with less then y_min samples are labeled as _other_ and are not used in the model.
-        10  (default)
-        None or 1
-
-    k : int, [1..n] , default : 1.
+        10, None, 1, etc
+    k : int, [1..n] , (default : 1)
         Number of combinatoric elements to create for the n features
-        The default is 1.
-
-    perc_min_num : float, Force column (int or float) to be numerical if unique non-zero values are above percentage.
-        None
-        0.8 (default)
-
-    multtest : String, default : 'holm'.
-         None            : No multiple Test
-        'bonferroni'     : one-step correction
-        'sidak'          : one-step correction
-        'holm-sidak'     : step down method using Sidak adjustments
-        'holm'           : step-down method using Bonferroni adjustments (default)
-        'simes-hochberg' : step-up method  (independent)
-        'hommel'         : closed method based on Simes tests (non-negative)
-        'fdr_bh'         : Benjamini/Hochberg  (non-negative)
-        'fdr_by'         : Benjamini/Yekutieli (negative)
-        'fdr_tsbh'       : two stage fdr correction (non-negative)
-        'fdr_tsbky'      : two stage fdr correction (non-negative)
-
-    dtypes : list of str, default : 'pandas'.
+    perc_min_num : float, (default : 0.8)
+        Force column (int or float) to be numerical if unique non-zero values are above percentage.
+        Can also be None.
+    multtest : String, (default : 'holm')
+        None: No multiple Test,
+        'bonferroni': one-step correction,
+        'sidak': one-step correction,
+        'holm-sidak': step down method using Sidak adjustments,
+        'holm': step-down method using Bonferroni adjustments,
+        'simes-hochberg': step-up method  (independent),
+        'hommel': closed method based on Simes tests (non-negative),
+        'fdr_bh': Benjamini/Hochberg  (non-negative),
+        'fdr_by': Benjamini/Yekutieli (negative),
+        'fdr_tsbh': two stage fdr correction (non-negative),
+        'fdr_tsbky': two stage fdr correction (non-negative)
+    dtypes : list of str, (default : 'pandas')
         list strings, example: ['cat','num',''] of length y. By default the dtype is determined based on the pandas dataframe. Empty ones [''] are skipped.
-        ['cat','cat','num','','cat',...]
-        'pandas' (default)
-
-    specificity : String, Configure how numerical data labels are stored. Setting this variable can be of use in the 'structure_learning' function for the creation of a network ([None] will glue most numerical labels together whereas [high] mostly will not).
-            None : No additional information in the labels
-            'low' : 'high' or 'low' are included that represents significantly higher or lower assocations compared to the rest-group.
-            'medium' : (default) 'high' or 'low' are included with 1 decimal behind the comma.
-            'high' : 'high' or 'low' are included with 3 decimal behind the comma.
-
-    dropna : Bool, [True,False] Drop rows/columns in adjacency matrix that showed no significance
-        True (default)
-        False
-
-    excl_background : String, String name to exclude from the background
-        None (default)
-        ['0.0']: To remove categorical values with label 0
-
+        Can also be of the form: ['cat','cat','num','','cat']
+    specificity : String, (default : 'medium')
+        Configure how numerical data labels are stored. Setting this variable can be of use in the 'association_learning' function for the creation of a network ([None] will glue most numerical labels together whereas [high] mostly will not).
+        None : No additional information in the labels,
+        'low' : 'high' or 'low' are included that represents significantly higher or lower assocations compared to the rest-group,
+        'medium': 'high' or 'low' are included with 1 decimal behind the comma,
+        'high' : 'high' or 'low' are included with 3 decimal behind the comma.
+    dropna : Bool, [True,False] (Default : True)
+        Drop rows/columns in adjacency matrix that showed no significance
+    excl_background : String (default : None)
+        Name to exclude from the background.
+        Example: '0.0': To remove categorical values with label 0
     black_list : List or None (default : None)
         If a list of edges is provided as black_list, they are excluded from the search and the resulting model will not contain any of those edges.
-
     white_list : List or None (default : None)
         If a list of edges is provided as white_list, the search is limited to those edges. The resulting model will then only contain edges that are in white_list.
 
@@ -137,7 +123,7 @@ class hnet():
     >>> hn = hnet()
     >>> # Load example dataset
     >>> df = hn.import_example('sprinkler')
-    >>> Structure learning
+    >>> association Learning
     >>> out = hn.association_learning(df)
     >>> # Plot dynamic graph
     >>> G_dynamic = hn.d3graph()
@@ -201,7 +187,7 @@ class hnet():
         # Print some
         if verbose>=3: print('[hnet] >Association learning across [%d] categories.' %(X_comb.shape[1]))
 
-        disable = (True if verbose==0 else False)
+        disable = (True if (verbose==0 or verbose>3) else False)
         count = 0
         nr_succes_pop_n = []
 
@@ -209,16 +195,17 @@ class hnet():
             [nr_succes_i, simmatP, simmat_labx] = _do_the_math(df, X_comb, dtypes, X_labx, simmatP, simmat_labx, i, self.specificity, self.y_min, verbose=verbose)
             nr_succes_pop_n.append(nr_succes_i)
             count = count + simmatP.shape[0]
+            if verbose>=4: print('[hnet] >[%d] %s' %(i, nr_succes_i))
 
         # Message
         if verbose>=3: print('[hnet] >Total number of computations: [%.0d]' %(count))
         # Post processing
-        [simmatP, simmatLogP, simmat_labx, nr_succes_pop_n] = hnstats._post_processing(simmatP, nr_succes_pop_n, simmat_labx, self.alpha, self.multtest, self.fillna, self.dropna, verbose=verbose)
+        simmatP, simmatLogP, simmat_labx, nr_succes_pop_n = hnstats._post_processing(simmatP, nr_succes_pop_n, simmat_labx, self.alpha, self.multtest, self.fillna, self.dropna, verbose=verbose)
         # Return
         return simmatP, simmatLogP, simmat_labx, nr_succes_pop_n
 
     def association_learning(self, df, verbose=3):
-        """Learn the structure in the data.
+        """Learn the associations in the data.
 
         Parameters
         ----------
@@ -258,10 +245,11 @@ class hnet():
         # Store
         self.results = _store(simmatP, simmatLogP, labx, df, nr_succes_pop_n, dtypes, rules)
         # Use this option for storage of your model
+        if verbose>=3: print('[hnet] >Fin.')
         return self.results
 
     # Make network d3
-    def d3graph(self, node_size_limits=[6, 15], savepath=None, node_color=None, directed=True, showfig=True):
+    def d3graph(self, node_size_limits=[6, 15], savepath=None, node_color=None, directed=True, threshold=None, white_list=None, black_list=None, min_edges=None, figsize=(1500, 1500), showfig=True, verbose=3):
         """Interactive network creator.
 
         Description
@@ -282,10 +270,20 @@ class hnet():
             Save the figure in specified path.
         node_color : None or 'cluster' default : None
             color nodes based on clustering or by label colors.
-        directed : bool, optional
-            Create network using directed edges (arrows). The default is True.
+        directed : bool, default is True.
+            Create network using directed edges (arrows).
+        threshold : int (default : None)
+            Associations (edges) are filtered based on the -log10(P) > threshold. threshold should range between 0 and maximum value of -log10(P).
+        black_list : List or None (default : None)
+            If a list of edges is provided as black_list, they are excluded from the search and the resulting model will not contain any of those edges.
+        white_list : List or None (default : None)
+            If a list of edges is provided as white_list, the search is limited to those edges. The resulting model will then only contain edges that are in white_list.
+        min_edges : int (default : None)
+            Edges are only shown if a node has at least min_edges.
         showfig : bool, optional
             Plot figure to screen. The default is True.
+        figsize : tuple, optional
+            Size of the figure in the browser, [height,width]. The default is [1500,1500].
 
         Returns
         -------
@@ -298,33 +296,44 @@ class hnet():
             labels of the nodes.
 
         """
+        status = self._check_results()
+        if not status: return None
+
+        if verbose>=3: print('[hnet] >Building d3graph..')
         # Setup tempdir
         savepath = hnstats._tempdir(savepath)
+        # Filter adjacency matrix on blacklist/whitelist and/or threshold
+        simmatLogP, labx = hnstats._filter_adjmat(self.results['simmatLogP'], self.results['labx'], threshold=threshold, min_edges=min_edges, white_list=white_list, black_list=black_list, verbose=verbose)
+        # Check whether anything has remained
+        if simmatLogP.values.flatten().sum()==0:
+            if verbose>=3: print('[hnet] >Nothing to plot.')
+            return None
 
-        [IA,IB] = ismember(self.results['simmatLogP'].columns, self.results['counts'][:,0])
-        node_size = np.repeat(node_size_limits[0], len(self.results['simmatLogP'].columns))
+        [IA,IB] = ismember(simmatLogP.columns, self.results['counts'][:,0])
+        node_size = np.repeat(node_size_limits[0], len(simmatLogP.columns))
         node_size[IA] = hnstats._scale_weights(self.results['counts'][IB,1], node_size_limits)
 
         # Make undirected network
-        if directed:
-            simmatLogP=self.results['simmatLogP']
-        else:
-            simmatLogP = to_undirected(self.results['simmatLogP'], method='logp')
+        if not directed:
+            simmatLogP = to_undirected(simmatLogP, method='logp')
 
         # Color node using network-clustering
         if node_color=='cluster':
-            labx = self.plot(node_color='cluster', showfig=False)['labx']
+            labx = self.plot(node_color='cluster', directed=True, threshold=threshold, white_list=white_list, black_list=black_list, min_edges=min_edges, showfig=False)['labx']
         else:
-            labx = label_encoder.fit_transform(self.results['labx'])
+            # IA,_ = ismember(self.results['simmatLogP'].columns, simmatLogP.columns)
+            # labx = label_encoder.fit_transform(self.results['labx'][IA])
+            labx = label_encoder.fit_transform(labx)
 
         # Make network
-        Gout = d3graphs(simmatLogP.T, savepath=savepath, node_size=node_size, charge=500, width=1500, height=800, collision=0.1, node_color=labx, directed=directed, showfig=showfig)
+        if verbose>=3: print('[hnet] >Creating output html..')
+        Gout = d3graphs(simmatLogP.T, savepath=savepath, node_size=node_size, charge=500, height=figsize[0], width=figsize[1], collision=0.1, node_color=labx, directed=directed, showfig=showfig)
         # Return
         Gout['labx'] = labx
         return(Gout)
 
     # Make network plot
-    def plot(self, scale=2, dist_between_nodes=0.4, node_size_limits=[25,500], node_color=None, savepath=None, figsize=[15,10], pos=None, layout='fruchterman_reingold', dpi=250, showfig=True):
+    def plot(self, scale=2, dist_between_nodes=0.4, node_size_limits=[25,500], directed=True, node_color=None, savepath=None, figsize=[15,10], pos=None, layout='fruchterman_reingold', dpi=250, threshold=None, white_list=None, black_list=None, min_edges=None, showfig=True, verbose=3):
         """Make plot static network plot of the model results.
 
         Description
@@ -343,6 +352,8 @@ class hnet():
             Nodes are scaled between the Min and max size. The default is [25,500].
         node_color : str, None or 'cluster' default is None
             color nodes based on clustering or by label colors.
+        directed : bool, default is True.
+            Create network using directed edges (arrows).
         savepath : str, optional
             Save the figure in specified path.
         figsize : tuple, optional
@@ -353,6 +364,14 @@ class hnet():
             layouts from networkx can be used. The default is 'fruchterman_reingold'.
         dpi : int, optional
             resolution of the figure. The default is 250.
+        threshold : int (default : None)
+            Associations (edges) are filtered based on the -log10(P) > threshold. threshold should range between 0 and maximum value of -log10(P).
+        black_list : List or None (default : None)
+            If a list of edges is provided as black_list, they are excluded from the search and the resulting model will not contain any of those edges.
+        white_list : List or None (default : None)
+            If a list of edges is provided as white_list, the search is limited to those edges. The resulting model will then only contain edges that are in white_list.
+        min_edges : int (default : None)
+            Edges are only shown if a node has at least min_edges.
         showfig : bool, optional
             Plot figure to screen. The default is True.
 
@@ -367,7 +386,11 @@ class hnet():
             Coordinates of the node postions.
 
         """
-        config = dict()
+        status = self._check_results()
+        if not status: return None
+        
+        if verbose>=3: print('[hnet] >Building network graph..')
+        config = {}
         config['scale'] = scale
         config['node_color'] = node_color
         config['dist_between_nodes'] = dist_between_nodes
@@ -378,9 +401,14 @@ class hnet():
 
         # Set savepath and filename
         config['savepath'] = hnstats._path_correct(savepath, filename='hnet_network', ext='.png')
-
         # Get adjmat
-        adjmatLog = self.results['simmatLogP'].copy()
+        # adjmatLog = self.results['simmatLogP'].copy()
+        # Filter adjacency matrix on blacklist/whitelist and/or threshold
+        adjmatLog, labx = hnstats._filter_adjmat(self.results['simmatLogP'], self.results['labx'], threshold=threshold, min_edges=min_edges, white_list=white_list, black_list=black_list, verbose=verbose)
+        # Check whether anything has remained
+        if adjmatLog.values.flatten().sum()==0:
+            if verbose>=3: print('[hnet] >Nothing to plot.')
+            return None
 
         # Set weights for edges
         adjmatLogWEIGHT = adjmatLog.copy()
@@ -388,14 +416,29 @@ class hnet():
         adjmatLogWEIGHT = pd.DataFrame(index=adjmatLog.index.values, data=MinMaxScaler(feature_range=(0, 20)).fit_transform(adjmatLogWEIGHT), columns=adjmatLog.columns)
 
         # Set size for node
-        [IA, IB] = ismember(self.results['simmatLogP'].columns, self.results['counts'][:, 0])
-        node_size = np.repeat(node_size_limits[0], len(self.results['simmatLogP'].columns))
+        [IA, IB] = ismember(adjmatLog.columns, self.results['counts'][:, 0])
+        node_size = np.repeat(node_size_limits[0], len(adjmatLog.columns))
         node_size[IA] = hnstats._scale_weights(self.results['counts'][IB, 1], node_size_limits)
 
         # Make new graph (G) and add properties to nodes
-        G = nx.DiGraph(directed=True)
+        if not directed:
+            adjmatLog = to_undirected(adjmatLog, method='logp')
+            G = nx.Graph()
+        else:
+            G = nx.DiGraph(directed=True)
+
+        # Color node using network-clustering
+        # if node_color=='cluster':
+            # _, labx = network.cluster(G.to_undirected())
+        # else:
+            # labx = self.results['labx']
+        # Make sure the labx are correct if previously black/white list filtering was performed
+        # IA,_ = ismember(self.results['simmatLogP'].columns, adjmatLog.columns)
+        # labx = label_encoder.fit_transform(labx[IA])
+        labx = label_encoder.fit_transform(labx)
+
         for i in range(0, adjmatLog.shape[0]):
-            G.add_node(adjmatLog.index.values[i], node_size=node_size[i], node_label=self.results['labx'][i])
+            G.add_node(adjmatLog.index.values[i], node_size=node_size[i], node_label=labx[i])
         # Add properties to edges
         np.fill_diagonal(adjmatLog.values, 0)
         for i in range(0, adjmatLog.shape[0]):
@@ -410,11 +453,11 @@ class hnet():
         edge_weights = np.array([G[u][v]['weight'] for u,v in edges])
         edge_weights = MinMaxScaler(feature_range=(0.5, 8)).fit_transform(edge_weights.reshape(-1, 1)).flatten()
 
-        # Cluster
+        # # Cluster
         if node_color=='cluster':
             _, labx = network.cluster(G.to_undirected())
         else:
-            labx = label_encoder.fit_transform(self.results['labx'])
+            labx = label_encoder.fit_transform(labx)
 
         # G = nx.DiGraph() # Directed graph
         # Layout
@@ -452,8 +495,8 @@ class hnet():
         Gout['pos'] = pos
         return(Gout)
 
-    # Make plot of the structure_learning
-    def heatmap(self, cluster=False, figsize=[15,10], savepath=None, verbose=3):
+    # Make plot of the association_learning
+    def heatmap(self, cluster=False, figsize=[15,10], savepath=None, threshold=None, white_list=None, black_list=None, min_edges=None, verbose=3):
         """Plot heatmap.
 
         Description
@@ -470,6 +513,14 @@ class hnet():
             Figure size. The default is [15, 10].
         savepath : Bool, optional
             saveingpath. The default is None.
+        threshold : int (default : None)
+            Associations (edges) are filtered based on the -log10(P) > threshold. threshold should range between 0 and maximum value of -log10(P).
+        black_list : List or None (default : None)
+            If a list of edges is provided as black_list, they are excluded from the search and the resulting model will not contain any of those edges.
+        white_list : List or None (default : None)
+            If a list of edges is provided as white_list, the search is limited to those edges. The resulting model will then only contain edges that are in white_list.
+        min_edges : int (default : None)
+            Edges are only shown if a node has at least min_edges.
         verbose : int, optional
             Verbosity. The default is 3.
 
@@ -478,7 +529,16 @@ class hnet():
         None.
 
         """
-        adjmatLog = self.results['simmatLogP'].copy()
+        status = self._check_results()
+        if not status: return None
+
+        # adjmatLog = self.results['simmatLogP'].copy()
+        # Filter adjacency matrix on blacklist/whitelist and/or threshold
+        adjmatLog, labx = hnstats._filter_adjmat(self.results['simmatLogP'], self.results['labx'], threshold=threshold, min_edges=min_edges, white_list=white_list, black_list=black_list, verbose=verbose)
+        if adjmatLog.values.flatten().sum()==0:
+            if verbose>=3: print('[hnet] >Nothing to plot.')
+            return None
+
         # Set savepath and filename
         savepath = hnstats._path_correct(savepath, filename='hnet_heatmap', ext='.png')
 
@@ -500,7 +560,7 @@ class hnet():
         except:
             print('[HNET.heatmap] Error: Heatmap failed. Try cluster=False')
 
-    # Extract combined rules from structure_learning
+    # Extract combined rules from association_learning
     def combined_rules(self, simmatP=None, labx=None, verbose=3):
         """Association testing and combining Pvalues using fishers-method.
 
@@ -567,23 +627,29 @@ class hnet():
         # Return
         return(df_rules)
 
-    def import_example(self, data='titanic', verbose=3):
+    def import_example(self, data='titanic', url=None, sep=',', verbose=3):
         """Import example dataset from github source.
-
+        
+        Description
+        -----------
+        Import one of the few datasets from github source or specify your own download url link.
+    
         Parameters
         ----------
-        data : str, optional
-            Name of the dataset 'sprinkler' or 'titanic' or 'student'.
-        verbose : int, optional
-            Print message to screen. The default is 3.
-
+        data : str
+            Name of datasets: 'sprinkler', 'titanic', 'student', 'fifa', 'cancer', 'waterpump', 'retail'
+        url : str
+            url link to to dataset.
+        verbose : int, (default: 3)
+            Print message to screen.
+    
         Returns
         -------
         pd.DataFrame()
             Dataset containing mixed features.
-
+    
         """
-        return import_example(data=data, verbose=verbose)
+        return import_example(data=data, url=url, sep=sep, verbose=verbose)
 
     # Save model
     def save(self, filepath='hnet_model.pkl', overwrite=False, verbose=3):
@@ -666,6 +732,18 @@ class hnet():
         else:
             if verbose>=2: print('[hnet] >WARNING: Could not load data.')
 
+    # Check results
+    def _check_results(self, verbose=3):
+        status = True
+        if not hasattr(self, 'results'):
+            if verbose>=3: print('[hnet] >Nothing to plot. Try to run hnet with first with: hn.association_learning()')
+            status = False
+        elif self.results['simmatLogP'].empty:
+            if verbose>=3: print('[hnet] >Nothing to plot. No associations were detected.')
+            status = False
+
+        return status
+
 
 # %% Store results
 def _store(simmatP, adjmatLog, labx, df, nr_succes_pop_n, dtypes, rules):
@@ -679,15 +757,21 @@ def _store(simmatP, adjmatLog, labx, df, nr_succes_pop_n, dtypes, rules):
     return out
 
 # %% Import example dataset from github.
-def import_example(data='titanic', verbose=3):
+def import_example(data='titanic', url=None, sep=',', verbose=3):
     """Import example dataset from github source.
+    
+    Description
+    -----------
+    Import one of the few datasets from github source or specify your own download url link.
 
     Parameters
     ----------
-    data : str, optional
-        Name of the dataset 'sprinkler' or 'titanic' or 'student'.
-    verbose : int, optional
-        Print message to screen. The default is 3.
+    data : str
+        Name of datasets: 'sprinkler', 'titanic', 'student', 'fifa', 'cancer', 'waterpump', 'retail'
+    url : str
+        url link to to dataset.
+    verbose : int, (default: 3)
+        Print message to screen.
 
     Returns
     -------
@@ -695,29 +779,46 @@ def import_example(data='titanic', verbose=3):
         Dataset containing mixed features.
 
     """
-    if data=='sprinkler':
-        url='https://erdogant.github.io/datasets/sprinkler.zip'
-    elif data=='titanic':
-        url='https://erdogant.github.io/datasets/titanic_train.zip'
-    elif data=='student':
-        url='https://erdogant.github.io/datasets/student_train.zip'
+    if url is None:
+        if data=='sprinkler':
+            url='https://erdogant.github.io/datasets/sprinkler.zip'
+        elif data=='titanic':
+            url='https://erdogant.github.io/datasets/titanic_train.zip'
+        elif data=='student':
+            url='https://erdogant.github.io/datasets/student_train.zip'
+        elif data=='cancer':
+            url='https://erdogant.github.io/datasets/cancer_dataset.zip'
+        elif data=='fifa':
+            url='https://erdogant.github.io/datasets/FIFA_2018.zip'
+        elif data=='waterpump':
+            url='https://erdogant.github.io/datasets/waterpump/waterpump_test.zip'
+        elif data=='retail':
+            url='https://erdogant.github.io/datasets/marketing_data_online_retail_small.zip'
+    else:
+        data = wget.filename_from_url(url)
+
+    if url is None:
+        if verbose>=3: print('[hnet] >Nothing to download.')
+        return None
 
     curpath = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
     PATH_TO_DATA = os.path.join(curpath, wget.filename_from_url(url))
+    if not os.path.isdir(curpath):
+        os.makedirs(curpath, exist_ok=True)
 
     # Check file exists.
     if not os.path.isfile(PATH_TO_DATA):
-        if verbose>=3: print('[hnet] >Downloading example dataset from github source..')
+        if verbose>=3: print('[hnet] >Downloading [%s] dataset from github source..' %(data))
         wget.download(url, curpath)
 
     # Import local dataset
     if verbose>=3: print('[hnet] >Import dataset [%s]' %(data))
-    df = pd.read_csv(PATH_TO_DATA)
+    df = pd.read_csv(PATH_TO_DATA, sep=sep)
     # Return
     return df
 
 # %% Compute fit
-def enrichment(df, y, y_min=None, alpha=0.05, multtest='holm', dtypes='pandas', specificity='medium', verbose=3):
+def enrichment(df, y, y_min=None, alpha=0.05, multtest='holm', dtypes='pandas', specificity='medium', excl_background=None, verbose=3):
     """Enrichment analysis.
 
     Description
@@ -741,6 +842,9 @@ def enrichment(df, y, y_min=None, alpha=0.05, multtest='holm', dtypes='pandas',
         By default the dtype is determined based on the pandas dataframe. Empty ones [''] are skipped. The default is 'pandas'.
     specificity : String, optional
         Configure how numerical data labels are stored.. The default is 'medium'.
+    excl_background : String (default : None)
+        Name to exclude from the background.
+        Example: ['0.0']: To remove categorical values with label 0
     verbose : int, optional
         Print message to screen. The higher the number, the more details. The default is 3.
 
@@ -787,7 +891,7 @@ def enrichment(df, y, y_min=None, alpha=0.05, multtest='holm', dtypes='pandas',
     assert 'numpy' in str(type(y)), 'y must be of type numpy array'
 
     # DECLARATIONS
-    config = dict()
+    config = {}
     config['verbose'] = verbose
     config['alpha'] = alpha
     config['multtest'] = multtest
@@ -795,13 +899,21 @@ def enrichment(df, y, y_min=None, alpha=0.05, multtest='holm', dtypes='pandas',
 
     if config['verbose']>=3: print('[hnet] >Start making fit..')
     df.columns = df.columns.astype(str)
+
+    # [df, df_onehot, dtypes] = hnstats._preprocessing(df, dtypes=dtypes, y_min=y_min, perc_min_num=perc_min_num, excl_background=excl_background, verbose=verbose)
+
+    df, dtypes, excl_background = hnstats._bool_processesing(df, dtypes, excl_background=excl_background, verbose=verbose)
+
+    # Convert bool columns to integer values
+    # if isinstance(dtypes, str):
+    #     Iloc = df.dtypes=='bool'
+    #     if np.any(Iloc):
+    #         df.loc[:,Iloc] = df.loc[:,Iloc].astype('int')
+
     # Set y as string
     y = df2onehot.set_y(y, y_min=y_min, verbose=config['verbose'])
     # Determine dtypes for columns
-    [df, dtypes] = df2onehot.set_dtypes(df, dtypes, verbose=config['verbose'])
-    
-    # [df, df_onehot, dtypes] = hnstats._preprocessing(df, dtypes=dtypes, y_min=y_min, perc_min_num=perc_min_num, excl_background=excl_background, verbose=verbose)
-
+    df, dtypes = df2onehot.set_dtypes(df, dtypes, verbose=config['verbose'])
     # Compute fit
     out = hnstats._compute_significance(df, y, dtypes, specificity=config['specificity'], verbose=config['verbose'])
     # Multiple test correction
@@ -843,6 +955,7 @@ def to_undirected(adjmat, method='logp', verbose=3):
     if np.sum(np.isin(adjmat.index.values, adjmat.columns.values))!=np.max(adjmat.shape):
         raise Exception('Adjacency matrix must have similar number of rows and columns! Re-run HNet with dropna=False!')
 
+    if verbose>=3: print('[hnet] >Make adjacency matrix undirected..')
     progressbar=(True if verbose==0 else False)
     columns=adjmat.columns.values
     index=adjmat.index.values
@@ -906,10 +1019,14 @@ def compare_networks(adjmat_true, adjmat_pred, pos=None, showfig=True, width=15,
     scores : dict
         Contains extensive number of keys with various scoring values.
     adjmat_diff : pd.DataFrame()
-        Differences between the two output matrices. Zero means no difference whereas value >0 does.
+        Difference of the input network compared to the second.
+        0 = No edge
+        1 = No difference between networks
+        2 = Addition of edge in the first input network compared to the second
+       -1 = Depliction of edge in the first netwwork compared to the second
 
     """
-    [scores, adjmat_diff] = network.compare_networks(adjmat_true,
+    scores, adjmat_diff = network.compare_networks(adjmat_true,
                                                      adjmat_pred,
                                                      pos=pos,
                                                      showfig=showfig,
@@ -921,13 +1038,14 @@ def compare_networks(adjmat_true, adjmat_pred, pos=None, showfig=True, width=15,
 
 # %% Do the math
 def _do_the_math(df, X_comb, dtypes, X_labx, simmatP, simmat_labx, i, specificity, y_min, verbose=3):
-    count=0
-    out=[]
+    count = 0
     # Get response variable to test association
-    y=X_comb.iloc[:,i].values.astype(str)
+    y = X_comb.iloc[:,i].values.astype(str)
     # Get column name
-    colname=X_comb.columns[i]
-    # Do something if response variable has more then 1 option
+    colname = X_comb.columns[i]
+    # Default output is nan
+    out = [colname, np.nan]
+    # Do math if response variable has more then 1 option
     if len(np.unique(y))>1:
         if verbose>=4: print('[hnet] >Working on [%s]' %(X_comb.columns[i]), end='')
         # Remove columns if it belongs to the same categorical subgroup; these can never overlap!
@@ -941,12 +1059,12 @@ def _do_the_math(df, X_comb, dtypes, X_labx, simmatP, simmat_labx, i, specificit
             # Column names
             idx = np.where(dfout['category_label'].isna())[0]
             catnames = dfout['category_name']
-            colnames = catnames + '_' + dfout['category_label']
+            colnames = catnames + '_' + dfout['category_label'].astype(str)
             colnames[idx] = catnames[idx].values
             # Add new column and index
-            [simmatP, simmat_labx] = hnstats._addcolumns(simmatP, colnames, simmat_labx, catnames)
+            simmatP, simmat_labx = hnstats._addcolumns(simmatP, colnames, simmat_labx, catnames)
             # Store values
-            [IA,IB]=ismember(simmatP.index.values.astype(str), colnames.values.astype(str))
+            IA, IB = ismember(simmatP.index.values.astype(str), colnames.values.astype(str))
             simmatP.loc[colname, IA] = dfout['Padj'].iloc[IB].values
             # Count nr. successes
             out = [colname, X_comb.iloc[:,i].sum() / X_comb.shape[0]]
