@@ -42,24 +42,29 @@ def _compute_significance(df, y, dtypes, specificity=None, verbose=3):
         for j in range(0, len(uiy)):
             target = uiy[j]
             # Catagorical
-            if dtypes[i]=='cat':
-                datacOnehot=pd.get_dummies(datac)
+            if (dtypes[i]=='cat'):
+                datacOnehot = pd.get_dummies(datac)
                 # Remove background column if 2 columns of which 0.0 also there
                 if (datacOnehot.shape[1]==2) & (np.any(np.isin(datacOnehot.columns, '0.0'))):
                     datacOnehot.drop(labels=['0.0'], axis=1, inplace=True)
+                # Remove False background column if 2 columns and False
+                if (datacOnehot.shape[1]==2) & (np.any(np.isin(datacOnehot.columns, False))):
+                    datacOnehot.drop(labels=[False], axis=1, inplace=True)
                 # Run over all unique entities/cats in column for target vlue
                 for k in range(0, datacOnehot.shape[1]):
                     outtest = _prob_hypergeo(datacOnehot.iloc[:, k], yc==target)
                     outtest.update({'y': target})
                     outtest.update({'category_name': colname})
                     out.append(outtest)
-
-            # Numerical
-            if dtypes[i]=='num':
+            elif dtypes[i]=='num':
+                # Numerical
                 outtest = _prob_ranksums(datac, yc==target, specificity=specificity)
                 outtest.update({'y': target})
                 outtest.update({'category_name': colname})
                 out.append(outtest)
+            else:
+                if verbose>=2: print('[hnet] >Warning: Can not process dtype: [%s] <skipping>' %(dtypes[i]))
+                # raise Exception('[hnet] >dtype can not be of type [%s]' %(dtypes[i]))
             # Print dots
             if verbose>=3: print('.', end='')
 
@@ -337,10 +342,10 @@ def _nancleaning(datac, y):
 # %% Do the math
 def _post_processing(simmat_padj, nr_succes_pop_n, simmat_labx, alpha, multtest, fillna, dropna, verbose=3):
     # Clean label names by chaning X.0 into X
-    simmat_padj.columns=list(map(lambda x: x[:-2] if x[-2:]=='.0' else x, simmat_padj.columns))
-    simmat_padj.index=list(map(lambda x: x[:-2] if x[-2:]=='.0' else x, simmat_padj.index.values))
-    nr_succes_pop_n=np.array(nr_succes_pop_n)
-    nr_succes_pop_n[:,0]=list(map(lambda x: x[:-2] if x[-2:]=='.0' else x, nr_succes_pop_n[:,0]))
+    simmat_padj.columns = list(map(lambda x: x[:-2] if x[-2:]=='.0' else x, simmat_padj.columns))
+    simmat_padj.index = list(map(lambda x: x[:-2] if x[-2:]=='.0' else x, simmat_padj.index.values))
+    nr_succes_pop_n = np.array(nr_succes_pop_n)
+    nr_succes_pop_n[:,0] = list(map(lambda x: x[:-2] if x[-2:]=='.0' else x, nr_succes_pop_n[:,0]))
 
     if verbose>=5: print(simmat_padj)
     # Multiple test correction
@@ -368,8 +373,8 @@ def _post_processing(simmat_padj, nr_succes_pop_n, simmat_labx, alpha, multtest,
         simmat_padj=simmat_padj.iloc[keepidx,keepidx]
         adjmatLog=adjmatLog.iloc[keepidx,keepidx]
         simmat_labx=simmat_labx[keepidx]
-        [IA,_]=ismember(nr_succes_pop_n[:,0], simmat_padj.columns.values)
-        nr_succes_pop_n=nr_succes_pop_n[IA,:]
+        IA, _ = ismember(nr_succes_pop_n[:,0], simmat_padj.columns.values)
+        nr_succes_pop_n = nr_succes_pop_n[IA,:]
 
     return(simmat_padj, adjmatLog, simmat_labx, nr_succes_pop_n)
 
@@ -420,18 +425,20 @@ def _path_correct(savepath, filename='fig', ext='.png'):
 
 
 # %%
-def _white_black_list(df, dtypes, white_list, black_list):
+def _white_black_list(df, dtypes, white_list, black_list, verbose=3):
     # Keep only variables that are in white_list.
     if white_list is not None:
         white_list = [x.lower() for x in white_list]
         Iloc = np.isin(df.columns.str.lower(), white_list)
         df = df.loc[:,Iloc]
+        if verbose>=3: print('[hnet] >Keeping ony features in the white list..')
 
     # Exclude variables that are in black_list.
     if black_list is not None:
         black_list = [x.lower() for x in black_list]
         Iloc = ~np.isin(df.columns.str.lower(), black_list)
         df = df.loc[:,Iloc]
+        if verbose>=3: print('[hnet] >Removing features from the black list..')
         # Remove also in dtypes
         if not isinstance(dtypes, str):
             if len(dtypes)!=len(Iloc):
@@ -445,23 +452,46 @@ def _white_black_list(df, dtypes, white_list, black_list):
             raise Exception('[hnet] >ERROR : dtypes should have same length as the white_list! Or do not specify dtypes.')
         if len(dtypes)!=df.shape[1]:
             raise Exception('[hnet] >ERROR : dtypes should have same length as your dataframe or white_list')
-    
-    if df.shape[1]<=1: raise Exception('[hnet] >ERROR : only %d variables are remaining. A minimum of 2 would be nice.' %(df.shape[1]))
+
+    if df.shape[1]<=1: print('[hnet] >Warning : After filtering, [%d] variable remained. A minimum of 2 is required. Tip: Check your dtypes, and see which ones are (not) categorical.' %(df.shape[1]))
     return df, dtypes
 
+# %%
+def _bool_processesing(df, dtypes, excl_background, verbose=3):
+    if isinstance(dtypes, str) | (np.any(dtypes=='bool')):
+        Iloc = ( (df.dtypes=='bool') | (dtypes=='bool') ).values
+        if np.any(Iloc):
+            if verbose>=3: print('[hnet] >Converting boolean values..')
+            # Set as int
+            df.loc[:,Iloc] = df.loc[:,Iloc].astype('int')
+            # Remove the background values
+            excl_background='0.0'
+            if verbose>=4: print('[hnet] >Set  parameter: excl_background=%s' %(str(excl_background)))
+            if not isinstance(dtypes, str):
+                # Set dtypes as catagorical
+                dtypes = np.array(dtypes)
+                dtypes[dtypes=='bool']='cat'
+    # Return
+    return df, dtypes, excl_background
 
 # %% Preprocessing
 def _preprocessing(df, dtypes='pandas', y_min=10, perc_min_num=0.8, excl_background=None, white_list=None, black_list=None, verbose=3):
+    if verbose>=4: print('[hnet] >preprocessing : Column names are set to str. and spaces are trimmed.')
     df.columns = df.columns.astype(str)
+    df.columns = df.columns.str.strip()
     df.reset_index(drop=True, inplace=True)
     
+    # Convert bool columns to integer values
+    df, dtypes, excl_background = _bool_processesing(df, dtypes, excl_background=excl_background, verbose=verbose)
     # Filter on white_list and black_list
     df, dtypes = _white_black_list(df, dtypes, white_list, black_list)
     # Remove columns without dtype
-    [df, dtypes] = _remove_columns_without_dtype(df, dtypes, verbose=verbose)
+    df, dtypes = _remove_columns_without_dtype(df, dtypes, verbose=verbose)
     # Make onehot matrix for response variable y
-    df_onehot = df2onehot.df2onehot(df, dtypes=dtypes, y_min=y_min, hot_only=True, perc_min_num=perc_min_num, excl_background=excl_background, verbose=verbose)
-    dtypes = df_onehot['dtypes']
+    df_onehot = df2onehot.df2onehot(df, dtypes=dtypes, y_min=y_min, deep_extract=False, hot_only=True, perc_min_num=perc_min_num, excl_background=excl_background, verbose=verbose)
+    # Set the dtypes for the input dataframe
+    df, dtypes = df2onehot.set_dtypes(df, dtypes=dtypes, deep_extract=False, perc_min_num=perc_min_num, num_if_decimal=True, verbose=0)
+    dtypes = np.array(dtypes)
 
     # Make sure its limited to the number of y_min
     Iloc = (df_onehot['onehot'].sum(axis=0)>=y_min).values
@@ -489,3 +519,46 @@ def _tempdir(savepath):
     os.makedirs(savepath, exist_ok=True)
 
     return(savepath)
+
+
+# %% Filter adjacency matrix
+def _filter_adjmat(simmatLogP, labx, threshold=None, min_edges=None, white_list=None, black_list=None, verbose=3):
+    # if (white_list is not None) and (len(white_list)<=1): raise Exception('[hnet] >ERROR: white_list should contain at least 2 input variables.')
+
+    # Filter on threshold
+    if threshold is not None:
+        if verbose>=3: print('[hnet] >Filtering associations on threshold > %d' %(threshold))
+        simmatLogP[simmatLogP<threshold]=0
+
+    # Filter on minimum number of edges
+    if min_edges is not None:
+        simmatBOOL = simmatLogP.copy()>0
+        if np.all(simmatBOOL.columns==simmatBOOL.index.values):
+            Iloc = np.logical_or(simmatBOOL.sum(axis=0)>=min_edges, simmatBOOL.sum(axis=1)>=min_edges).values
+            if verbose>=3: print('[hnet] >Filtering on edges: [%d] variables remain after filtering on a minimum of [%d] edges.' %(np.sum(Iloc), min_edges))
+            simmatLogP = simmatLogP.loc[Iloc,Iloc]
+            labx = labx[Iloc]
+
+    # Filter on white_list
+    if white_list is not None:
+        Irow = np.isin(simmatLogP.index.values, white_list)
+        Icol = np.isin(simmatLogP.columns, white_list)
+        Ilabx = np.isin(labx, white_list)
+        Irow = np.logical_or(Irow, Ilabx)
+        Icol = np.logical_or(Icol, Ilabx)
+        if verbose>=3: print('[hnet] >Number of variables at input: [%d], and after white listing: [%d]' %(len(Irow), sum(Irow)))
+        simmatLogP = simmatLogP.iloc[Irow,Icol]
+        labx = labx[Irow]
+
+    # Filter on white_list
+    if black_list is not None:
+        Irow = np.isin(simmatLogP.index.values, black_list)
+        Icol = np.isin(simmatLogP.columns, black_list)
+        Ilabx = np.isin(labx, black_list)
+        Irow = np.logical_or(Irow, Ilabx)
+        Icol = np.logical_or(Icol, Ilabx)
+        if verbose>=3: print('[hnet] >Number of variables at input: [%d], and after black listing: [%d]' %(len(Irow), sum(~Irow)))
+        simmatLogP = simmatLogP.iloc[~Irow, ~Icol]
+        labx = labx[~Irow]
+
+    return simmatLogP, labx
