@@ -81,7 +81,7 @@
 import matplotlib.pyplot as plt
 import pandas as pd
 import numpy as np
-import community
+from community import community_louvain
 import networkx as nx
 from sklearn.preprocessing import minmax_scale
 from sklearn.preprocessing import MinMaxScaler 
@@ -172,7 +172,7 @@ def compute_centrality(G, centrality='betweenness', verbose=3):
 def cluster(G, verbose=3):
     if verbose>=3: print('[NETWORK.cluster] Clustering using best partition')
     # Partition
-    partition=community.best_partition(G)
+    partition=community_louvain.best_partition(G)
     # Set property to node
     nx.set_node_attributes(G, partition, 'clusterlabel')
     # Extract labels
@@ -388,14 +388,14 @@ def compare_networks(adjmat_true, adjmat_pred, pos=None, showfig=True, width=15,
     adjmat_true = adjmat_true.loc[IArow,IAcol]
     adjmat_pred = adjmat_pred.iloc[IBrow,IBcol]
 
-    # Make sure it is boolean adjmat
-    adjmat_true = adjmat_true>0
-    adjmat_pred = adjmat_pred>0
-
     # Check whether order is correct
     if not np.all(adjmat_true.columns.values==adjmat_pred.columns.values): raise Exception('Column order of both input values could not be matched')
     if not np.all(adjmat_true.index.values==adjmat_pred.index.values): raise Exception('Row order of both input values could not be matched')
 
+    # Make sure it is boolean adjmat
+    adjmat_true = adjmat_true>0
+    adjmat_pred = adjmat_pred>0
+
     # Get edges
     y_true = adjmat_true.stack().reset_index()[0].values
     y_pred = adjmat_pred.stack().reset_index()[0].values
@@ -405,6 +405,7 @@ def compare_networks(adjmat_true, adjmat_pred, pos=None, showfig=True, width=15,
     #bayes.plot(out_bayes['adjmat'], pos=G['pos'])
 
     # Setup graph
+    # adjmat_diff = adjmat_true - adjmat_pred
     adjmat_diff = adjmat_true.astype(int)
     adjmat_diff[(adjmat_true.astype(int) - adjmat_pred.astype(int))<0]=2
     adjmat_diff[(adjmat_true.astype(int) - adjmat_pred.astype(int))>0]=-1
